(* Demonstration of double-buffer rendering.

   We render to a pixmap and then copy the contents of the pixmap to the
   window with xCopyArea().

   This will be similar to many implementations of glXSwapBuffers, however,
   many GL implementation allow synchronization with the refresh rate,
   while Xlib offers no such feature.
   Note that glXSwapBuffers executes an implicit glFlush, while xCopyArea
   does not execute an implicit xFlush.
*)
open Xlib

let () =
  let width = 500 and height = 500 in
  let dpy = xOpenDisplay "" in

  (* initialisation of the standard variables *)
  let screen = xDefaultScreen dpy in
  let root = xDefaultRootWindow dpy
  and visual = xDefaultVisual dpy screen
  and depth = xDefaultDepth dpy screen
  and fg = xBlackPixel dpy screen
  and bg = xWhitePixel dpy screen
  in

  (* creation of the double buffer *)
  let _root = drawable_of_window root in
  let db = xCreatePixmap dpy _root width height depth in

  let gcvalues = new_xGCValues() in

  xGCValues_set_foreground gcvalues fg;
  xGCValues_set_background gcvalues bg;

  let gc = xCreateGC dpy _root [GCForeground; GCBackground] gcvalues in

  (* window attributes *)
  let xswa = new_xSetWindowAttributes() in

  (* the events we want *)
  winAttr_set_event_mask xswa [ExposureMask;ButtonPressMask;ButtonReleaseMask;
                               PointerMotionMask;KeyPressMask];

  (* border and background colors *)
  winAttr_set_background_pixel xswa bg;
  winAttr_set_border_pixel xswa fg;

  let win = xCreateWindow dpy root
                          100 100 width height 2
                          depth InputOutput visual
                          [CWEventMask;CWBorderPixel;CWBackPixel]
                          xswa in

  xMapWindow dpy win;

  let _db = drawable_of_pixmap db
  and _win = drawable_of_window win in

  let report = new_xEvent() in
  while true do
    xNextEvent dpy report;
    match xEventType report with
    | Expose ->
        (* remove all the Expose events from the event stack *)
        while (xCheckTypedEvent dpy Expose report) do () done;
        xCopyArea dpy _db _win gc 0 0 width height 0 0;
        (* force refresh the screen *)
        xFlush dpy;

    | ButtonPress ->
        let xbutton = xButtonEvent_datas(to_xButtonEvent report) in
        begin match xbutton.button with
        (* left clic *)
        | Button1 ->
            (* animation with the double buffer *)
            for j=0 to pred 100 do
              xSetForeground dpy gc bg;
              xFillRectangle dpy _db gc 0 0 width height;
              xSetForeground dpy gc fg;

              for i=0 to pred 100 do
                let i = i  * 4 in
                xDrawArc dpy _db gc (xbutton.button_x + j)
                                    (xbutton.button_y + j) i i 0 (360 * 64);
              done;

              xCopyArea dpy _db _win gc 0 0 width height 0 0;
              (* force refresh the screen *)
              xFlush dpy;
            done;

        (* right and middle clic *)
        | Button2
        | Button3 ->
            (* animation without double buffer *)
            for j=0 to pred 100 do
              xClearWindow dpy win;

              for i=0 to pred 100 do
                let i = i * 4 in
                xDrawArc dpy _win gc (xbutton.button_x + j)
                                     (xbutton.button_y + j) i i  0 (360 * 64);
              done;
            done;
        | _ -> ()
        end;

    | KeyPress ->
        (* exit on any key press *)
        xCloseDisplay dpy;
        exit 0;

    | ConfigureNotify -> ()
    | _ -> ()
  done;
;;

