/* OCaml bindings for the GLX library (as part of OCaml-Xlib).
 * Copyright (C) 2008 by Florent Monnier <fmonnier@linux-nantes.org>
 *
 * OCaml-Xlib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * OCaml-Xlib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with OCaml-Xlib.  If not, see:
 * <http://www.gnu.org/licenses/>
 */

#include "wrap_xlib.h"

#include <GL/glx.h>

#define CAML_NAME_SPACE 1

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/fail.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// pointers to structures:

#define GLXContext_val(v) ((GLXContext)(v))
#define Val_GLXContext(c) ((value)(c))

#define GLXFBConfig_val(v) ((GLXFBConfig)(v))
#define Val_GLXFBConfig(c) ((value)(c))

// XID's

#define GLXDrawable_val(v) (XID_val(GLXDrawable,(v)))
#define Val_GLXDrawable(xid) (Val_XID((xid)))

#define GLXWindow_val(v) (XID_val(GLXWindow,(v)))
#define Val_GLXWindow(xid) (Val_XID((xid)))

#define GLXPbuffer_val(v) (XID_val(GLXPbuffer,(v)))
#define Val_GLXPbuffer(xid) (Val_XID((xid)))


CAMLprim value
ml_glXQueryExtension_tpl( value dpy )
{
    CAMLparam1(dpy);
    CAMLlocal1(tpl);
    int error_base, event_base;
    Bool ret = glXQueryExtension( Display_val(dpy), &error_base, &event_base );
    tpl = caml_alloc(3, 0);
    Store_field( tpl, 0, Val_bool(ret) );
    Store_field( tpl, 1, Val_int(error_base) );
    Store_field( tpl, 2, Val_int(event_base) );
    CAMLreturn(tpl);
}

CAMLprim value
ml_glXQueryExtension( value dpy )
{
    int error_base, event_base;
    if (!glXQueryExtension( Display_val(dpy), &error_base, &event_base )) {
        fprintf(stderr, "error_base: %d;  event_base: %d\n",
                         error_base, event_base);
        fflush(stderr);
        caml_failwith("OpenGL GLX extension not supported by display");
    }
    return Val_unit;
}

CAMLprim value
ml_glXQueryVersion( value dpy )
{
    CAMLparam1(dpy);
    CAMLlocal1(tpl);
    int major, minor;
    if (!glXQueryVersion( Display_val(dpy), &major, &minor ))
        caml_failwith("glXQueryVersion");
    tpl = caml_alloc(2, 0);
    Store_field( tpl, 0, Val_int(major) );
    Store_field( tpl, 1, Val_int(minor) );
    CAMLreturn(tpl);
}

CAMLprim value
ml_glXGetClientString( value dpy, value name )
{
    const int name_table[] = {
        GLX_VENDOR,
        GLX_VERSION,
        GLX_EXTENSIONS
    };
    const char *str = glXGetClientString( Display_val(dpy), name_table[Long_val(name)] );
    return caml_copy_string(str);
}

CAMLprim value
ml_glXCreateWindow( value dpy, value config, value win )
{
    int attrib_list[] = { None };  // the man says: currently unused
    GLXWindow glx_win = glXCreateWindow( Display_val(dpy), GLXFBConfig_val(config),
                                         Window_val(win), attrib_list );
    return Val_GLXWindow(glx_win);
}

CAMLprim value
ml_glXDestroyWindow( value dpy, value win )
{
    glXDestroyWindow( Display_val(dpy), GLXWindow_val(win) );
    return Val_unit;
}

CAMLprim value
ml_glXSwapBuffers( value dpy, value drawable )
{
    glXSwapBuffers( Display_val(dpy), GLXDrawable_val(drawable) );
    return Val_unit;
}

CAMLprim value
ml_glXWaitGL(value unit)
{
    glXWaitGL();
    return Val_unit;
}

CAMLprim value
ml_glXWaitX(value unit)
{
    glXWaitX();
    return Val_unit;
}

static inline int
visual_attrib( value v, int *p1, int *p2 )
{
    if (Is_long(v))
    {
        switch (Int_val(v))
        {
            case 0: *p1 = GLX_USE_GL; break;
            case 1: *p1 = GLX_RGBA; break;
            case 2: *p1 = GLX_DOUBLEBUFFER; break;
            case 3: *p1 = GLX_STEREO; break;
            default: caml_failwith("visual_attrib handling bug");
        }
        return 1;
    }
    else // (Is_block(v))
    {
        switch (Tag_val(v))
        {
            case 0:  *p1 = GLX_BUFFER_SIZE;      *p2 = UInt_val(Field(v,0)); break;
            case 1:  *p1 = GLX_LEVEL;            *p2 = Int_val(Field(v,0)); break;
            case 2:  *p1 = GLX_AUX_BUFFERS;      *p2 = UInt_val(Field(v,0)); break;
            case 3:  *p1 = GLX_RED_SIZE;         *p2 = UInt_val(Field(v,0)); break;
            case 4:  *p1 = GLX_GREEN_SIZE;       *p2 = UInt_val(Field(v,0)); break;
            case 5:  *p1 = GLX_BLUE_SIZE;        *p2 = UInt_val(Field(v,0)); break;
            case 6:  *p1 = GLX_ALPHA_SIZE;       *p2 = UInt_val(Field(v,0)); break;
            case 7:  *p1 = GLX_DEPTH_SIZE;       *p2 = UInt_val(Field(v,0)); break;
            case 8:  *p1 = GLX_STENCIL_SIZE;     *p2 = UInt_val(Field(v,0)); break;
            case 9:  *p1 = GLX_ACCUM_RED_SIZE;   *p2 = UInt_val(Field(v,0)); break;
            case 10: *p1 = GLX_ACCUM_GREEN_SIZE; *p2 = UInt_val(Field(v,0)); break;
            case 11: *p1 = GLX_ACCUM_BLUE_SIZE;  *p2 = UInt_val(Field(v,0)); break;
            case 12: *p1 = GLX_ACCUM_ALPHA_SIZE; *p2 = Int_val(Field(v,0)); break;
            default: caml_failwith("visual_attrib handling bug");
        }
        return 2;
    }
}

CAMLprim value
ml_glXChooseVisual( value dpy, value screen, value ml_attribList )
{
#define MAX_ATTRIBS 31
    int attribList[MAX_ATTRIBS];
    XVisualInfo *visinfo;

    int i = 0;
    while ( ml_attribList != Val_emptylist )
    {
        int n, p1, p2;
        value attrib = Field(ml_attribList, 0);
        n = visual_attrib( attrib, &p1, &p2 );
        if (n == 1) {
            if (i > MAX_ATTRIBS - 2)  // i max = 29
                caml_invalid_argument("visual_attrib list: too much elements");
            attribList[i] = p1;
            ++i;
        } else {
            if (i > MAX_ATTRIBS - 3)  // i max = 28
                caml_invalid_argument("visual_attrib list: too much elements");
            attribList[i] = p1;
            attribList[i+1] = p2;
            i += 2;
        }
        ml_attribList = Field(ml_attribList, 1);
    }
    attribList[i] = None;  // i max = 30

    visinfo = glXChooseVisual( Display_val(dpy), ScreenNB_val(screen), attribList );
    if (!visinfo) caml_failwith("glXChooseVisual");
    return Val_XVisualInfo(visinfo);
}

CAMLprim value
ml_glXCreateContext( value dpy, value vis, value share_list, value direct )
{
    GLXContext shareList;

    if (share_list == Val_int(0)) shareList = NULL;  // None
    else shareList = GLXContext_val( Field(share_list,0) );  // Some v

    GLXContext ctx = glXCreateContext( Display_val(dpy), XVisualInfo_val(vis),
                                       shareList, Bool_val(direct) );

    if (!ctx) caml_failwith("glXCreateContext");

    return Val_GLXContext(ctx);
}

CAMLprim value
ml_glXCopyContext( value dpy, value src, value mask )
{
    GLXContext dst;
    glXCopyContext( Display_val(dpy), GLXContext_val(src), dst, Long_val( mask ) );
    if (!dst) caml_failwith("glXCopyContext");
    return Val_GLXContext(dst);
}

CAMLprim value
ml_glXDestroyContext( value dpy, value ctx )
{
    glXDestroyContext( Display_val(dpy), GLXContext_val(ctx) );
    return Val_unit;
}

CAMLprim value
ml_glXMakeCurrent( value dpy, value drawable, value ctx )
{
    if (!glXMakeCurrent( Display_val(dpy), GLXDrawable_val(drawable), GLXContext_val(ctx) ))
        caml_failwith("glXMakeCurrent");
    return Val_unit;
}

CAMLprim value
ml_glXIsDirect( value dpy, value ctx )
{
    Bool b = glXIsDirect( Display_val(dpy), GLXContext_val(ctx) );
    return Val_bool(b);
}

CAMLprim value
ml_glXCreatePbuffer( value dpy, value config, value attrib_list )
{
    int attrs[20];
    int i = 0;
    while ( attrib_list != Val_emptylist )
    {
        value attrib = Field(attrib_list, 0);
        {
            switch (Tag_val(attrib))
            {
                case 0: attrs[i++] = GLX_PBUFFER_WIDTH; break;
                case 1: attrs[i++] = GLX_PBUFFER_HEIGHT; break;
                case 2: attrs[i++] = GLX_LARGEST_PBUFFER; break;
                case 3: attrs[i++] = GLX_PRESERVED_CONTENTS; break;
                default: caml_failwith("variant handling bug");
            }
            attrs[i++] = Int_val(Field(attrib,0));
        }
        attrib_list = Field(attrib_list,1);
        if (i >= 19) break;
    }
    attrs[i] = None;

    GLXPbuffer pbuf = glXCreatePbuffer( Display_val(dpy), GLXFBConfig_val(config), attrs );
    return Val_GLXPbuffer(pbuf);
}

CAMLprim value
ml_glXDestroyPbuffer( value dpy, value pbuf )
{
    glXDestroyPbuffer( Display_val(dpy), GLXPbuffer_val(pbuf) );
    return Val_unit;
}

// vim: sw=4 sts=4 ts=4 et fdm=marker
