/* OCaml bindings for the Xlib library.
 * Copyright (C) 2008 by Florent Monnier <fmonnier@linux-nantes.org>
 *
 * OCaml-Xlib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * OCaml-Xlib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with OCaml-Xlib.  If not, see:
 * <http://www.gnu.org/licenses/>
 */

// {{{ Headers 

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#define CAML_NAME_SPACE 1

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/fail.h>

// }}}

#include "wrap_xlib.h"

// {{{ XID's 

#define Drawable_val(v) (XID_val(Drawable,(v)))
#define Val_Drawable(xid) (Val_XID((xid)))

#define Font_val(v) (XID_val(Font,(v)))
#define Val_Font(xid) (Val_XID((xid)))

#define Pixmap_val(v) (XID_val(Pixmap,(v)))
#define Val_Pixmap(xid) (Val_XID((xid)))

#define Cursor_val(v) (XID_val(Cursor,(v)))
#define Val_Cursor(xid) (Val_XID((xid)))

#define Colormap_val(v) (XID_val(Colormap,(v)))
#define Val_Colormap(xid) (Val_XID((xid)))

#define GContext_val(v) (XID_val(GContext,(v)))
#define Val_GContext(xid) (Val_XID((xid)))

//#define KeySym_val(v) (XID_val(KeySym,(v)))
//#define Val_KeySym(xid) (Val_XID((xid)))
/* Get keysyms values instead of an abstract type */
#define Val_keysym Val_int
#define Keysym_val Long_val

// }}}
// {{{ Pointers 

// the type GC is a pointer to a structure (so no star)
#define GC_val(v) ((GC) (v))
#define Val_GC(gc) ((value)(gc))

// }}}
// {{{ Event-Masks 

static const long event_mask_table[] = {
    KeyPressMask,
    KeyReleaseMask,
    ButtonPressMask,
    ButtonReleaseMask,
    EnterWindowMask,
    LeaveWindowMask,
    PointerMotionMask,
    PointerMotionHintMask,
    Button1MotionMask,
    Button2MotionMask,
    Button3MotionMask,
    Button4MotionMask,
    Button5MotionMask,
    ButtonMotionMask,
    KeymapStateMask,
    ExposureMask,
    VisibilityChangeMask,
    StructureNotifyMask,
    ResizeRedirectMask,
    SubstructureNotifyMask,
    SubstructureRedirectMask,
    FocusChangeMask,
    PropertyChangeMask,
    ColormapChangeMask,
    OwnerGrabButtonMask,
};

static inline long
Eventmask_val( value em_list )
{
    long event_mask = 0;
    while ( em_list != Val_emptylist )
    {
        value head = Field(em_list, 0);
        long mask = event_mask_table[Long_val(head)];
        event_mask |= mask;
        em_list = Field(em_list, 1);
    }
    return event_mask;
}

// }}}
// {{{ caml allocs 

custom_ops(XEvent);
custom_ops(XColor);
custom_ops(XGCValues);
custom_ops(XSetWindowAttributes);
custom_ops(XSizeHints);

#define copy_XEvent(e,v) copy_obj(XEvent,(e),(v))

#define alloc_XEvent(v) alloc_obj(XEvent,(v))
#define alloc_XColor(v) alloc_obj(XColor,(v))
#define alloc_XGCValues(v) alloc_obj(XGCValues,(v))
#define alloc_XSetWindowAttributes(v) alloc_obj(XSetWindowAttributes,(v))
#define alloc_XSizeHints(v) alloc_obj(XSizeHints,(v))

#define XEvent_val(v) ((XEvent *)Data_custom_val(v))
#define XColor_val(v) ((XColor *)Data_custom_val(v))
#define XGCValues_val(v) ((XGCValues *)Data_custom_val(v))
#define XSetWindowAttributes_val(v) ((XSetWindowAttributes *)Data_custom_val(v))
#define XSizeHints_val(v) ((XSizeHints *)Data_custom_val(v))

// }}}
// {{{ macro/funcs 

/* switch this to use the macros or the functions */
#define XMP(f) X##f  // use the functions
//#define XMP(f) f  // use the macros


#define _xConnectionNumber XMP(ConnectionNumber)
#define _xRootWindow XMP(RootWindow)
#define _xDefaultScreen XMP(DefaultScreen)
#define _xDefaultRootWindow XMP(DefaultRootWindow)
#define _xDefaultVisual XMP(DefaultVisual)
#define _xDefaultGC XMP(DefaultGC)
#define _xBlackPixel XMP(BlackPixel)
#define _xWhitePixel XMP(WhitePixel)

#define _xDisplayWidth XMP(DisplayWidth)
#define _xDisplayHeight XMP(DisplayHeight)

#define _xScreenCount      XMP(ScreenCount)
#define _xServerVendor     XMP(ServerVendor)
#define _xProtocolVersion  XMP(ProtocolVersion)
#define _xProtocolRevision XMP(ProtocolRevision)
#define _xVendorRelease    XMP(VendorRelease)
#define _xDisplayString    XMP(DisplayString)
#define _xDefaultDepth     XMP(DefaultDepth)
#define _xDefaultColormap  XMP(DefaultColormap)
#define _xBitmapUnit       XMP(BitmapUnit)
#define _xBitmapBitOrder   XMP(BitmapBitOrder)
#define _xBitmapPad        XMP(BitmapPad)
#define _xImageByteOrder   XMP(ImageByteOrder)



#define _xScreenOfDisplay         XMP(ScreenOfDisplay)
#define _xDefaultScreenOfDisplay  XMP(DefaultScreenOfDisplay)
#define _xDisplayOfScreen         XMP(DisplayOfScreen)
#define _xRootWindowOfScreen      XMP(RootWindowOfScreen)
#define _xBlackPixelOfScreen      XMP(BlackPixelOfScreen)
#define _xWhitePixelOfScreen      XMP(WhitePixelOfScreen)
#define _xDefaultColormapOfScreen XMP(DefaultColormapOfScreen)
#define _xDefaultDepthOfScreen    XMP(DefaultDepthOfScreen)
#define _xDefaultGCOfScreen       XMP(DefaultGCOfScreen)
#define _xDefaultVisualOfScreen   XMP(DefaultVisualOfScreen)
#define _xWidthOfScreen           XMP(WidthOfScreen)
#define _xHeightOfScreen          XMP(HeightOfScreen)
#define _xWidthMMOfScreen         XMP(WidthMMOfScreen)
#define _xHeightMMOfScreen        XMP(HeightMMOfScreen)
#define _xPlanesOfScreen          XMP(PlanesOfScreen)
#define _xCellsOfScreen           XMP(CellsOfScreen)
#define _xMinCmapsOfScreen        XMP(MinCmapsOfScreen)
#define _xMaxCmapsOfScreen        XMP(MaxCmapsOfScreen)
#define _xDoesSaveUnders          XMP(DoesSaveUnders)
#define _xDoesBackingStore        XMP(DoesBackingStore)
#define _xEventMaskOfScreen       XMP(EventMaskOfScreen)


// }}}
// {{{ ints 

// wraps unsigned long
#define Pixel_color_val Unsigned_long_val
#define Val_pixel_color Val_long

// }}}
// {{{ return status 

#define DO_CHECK_RETURN_STATUS 1
#if DO_CHECK_RETURN_STATUS

#define GET_STATUS  int status =
#define CHECK_STATUS(fun,st) \
    { if (status != st) caml_failwith(#fun ": wrong return status"); }

#else

#define GET_STATUS
#define CHECK_STATUS

#endif

// }}}

CAMLprim value
ml_XlibSpecificationRelease( value unit )
{
    return Val_int(XlibSpecificationRelease);
}

CAMLprim value
ml_XOpenDisplay( value display_name )
{
    Display *dpy;
    dpy = XOpenDisplay( String_val(display_name) );
    if (dpy == NULL) {
        caml_failwith("Cannot open display");
    }
    return Val_Display(dpy);
}

CAMLprim value
ml_XCloseDisplay( value dpy )
{
    XCloseDisplay( Display_val(dpy) );
    return Val_unit;
}

CAMLprim value
ml_XFlush( value dpy )
{
    GET_STATUS  XFlush( Display_val(dpy) );
    CHECK_STATUS(XFlush, 1);
    return Val_unit;
}

CAMLprim value
ml_XDefaultScreen( value dpy )
{
    int screen_number = _xDefaultScreen( Display_val(dpy) );
    return Val_screenNB(screen_number);
}

CAMLprim value
ml_XDefaultRootWindow( value dpy )
{
    return Val_Window( _xDefaultRootWindow( Display_val(dpy) ));
}

CAMLprim value
ml_XDefaultVisual( value dpy, value screen_number )
{
    Visual * vis = _xDefaultVisual(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_Visual(vis);
}

CAMLprim value
ml_XDefaultDepth( value dpy, value screen_number )
{
    int depth = _xDefaultDepth(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_int(depth);
}

CAMLprim value
ml_XProtocolVersion( value dpy )
{
    return Val_int( _xProtocolVersion( Display_val(dpy) ));
}

CAMLprim value
ml_XProtocolRevision( value dpy )
{
    return Val_int( _xProtocolRevision( Display_val(dpy) ));
}

CAMLprim value
ml_XVendorRelease( value dpy )
{
    return Val_int( _xVendorRelease( Display_val(dpy) ));
}

CAMLprim value
ml_XServerVendor( value dpy )
{
    char * vendor = _xServerVendor( Display_val(dpy) );
    return caml_copy_string(vendor);
}

CAMLprim value
ml_XBlackPixel( value dpy, value screen_number )
{
    unsigned long color = _xBlackPixel(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_pixel_color(color);
}

CAMLprim value
ml_XWhitePixel( value dpy, value screen_number )
{
    unsigned long color = _xWhitePixel(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_pixel_color(color);
}

CAMLprim value
ml_XDisplayWidth( value dpy, value screen_number )
{
    int width = _xDisplayWidth(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_int(width);
}

CAMLprim value
ml_XDisplayHeight( value dpy, value screen_number )
{
    int height = _xDisplayHeight(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_int(height);
}

CAMLprim value
ml_XRootWindow( value dpy, value screen_number )
{
    Window win = _xRootWindow(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_Window(win);
}

CAMLprim value
ml_XDefaultGC( value dpy, value screen_number )
{
    GC gc = _xDefaultGC(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_GC(gc);
}

/* {{{ XColor */

CAMLprim value
ml_alloc_XColor( value unit )
{
    CAMLparam0();
    CAMLlocal1(x_color);
    alloc_XColor(x_color);
    memset(XColor_val(x_color), 0, sizeof(XColor));
    CAMLreturn(x_color);
}

CAMLprim value
ml_XColor_set_red( value x_color, value v )
{
    XColor * xcolor = XColor_val(x_color);
    xcolor->red = Long_val(v);
    return Val_unit;
}

CAMLprim value
ml_XColor_set_green( value x_color, value v )
{
    XColor * xcolor = XColor_val(x_color);
    xcolor->green = Long_val(v);
    return Val_unit;
}

CAMLprim value
ml_XColor_set_blue( value x_color, value v )
{
    XColor * xcolor = XColor_val(x_color);
    xcolor->blue = Long_val(v);
    return Val_unit;
}

static const char color_flags_table[] = {
    DoRed,
    DoGreen,
    DoBlue,
};

CAMLprim value
ml_XColor_set_flags( value x_color, value mask_list )
{
    XColor * xcolor = XColor_val(x_color);

    while ( mask_list != Val_emptylist )
    {
        value head = Field(mask_list, 0);
        xcolor->flags |= color_flags_table[Long_val(head)];
        mask_list = Field(mask_list, 1);
    }
    return Val_unit;
}

CAMLprim value
ml_XAllocColor( value dpy, value colormap, value x_color )
{
    XColor * xcolor = XColor_val(x_color);
    XAllocColor( Display_val(dpy), Colormap_val(colormap), xcolor );
    return Val_unit;
}

CAMLprim value
ml_XColor_pixel( value x_color )
{
    XColor * xcolor = XColor_val(x_color);
    return Val_pixel_color(xcolor->pixel);
}

/* }}} */

CAMLprim value
ml_XCreateSimpleWindow( value dpy, value parent, value x, value y,
                        value width, value height, value border_width,
                        value border, value background)
{
    Window win = XCreateSimpleWindow(
        Display_val(dpy),
        Window_val(parent),
        Int_val(x),
        Int_val(y),
        UInt_val(width),
        UInt_val(height),
        UInt_val(border_width),
        Pixel_color_val(border),
        Pixel_color_val(background) );
    return Val_Window(win);
}
CAMLprim value
ml_XCreateSimpleWindow_bytecode( value * argv, int argn )
{
    return ml_XCreateSimpleWindow( argv[0], argv[1], argv[2], argv[3],
                                   argv[4], argv[5], argv[6], argv[7], argv[8] );
}

CAMLprim value
ml_XDestroyWindow( value dpy, value win )
{
    GET_STATUS  XDestroyWindow(
        Display_val(dpy),
        Window_val(win) );
    CHECK_STATUS(XDestroyWindow, 1);
    return Val_unit;
}

CAMLprim value
ml_XVisualInfo_contents( value visual_info )
{
    CAMLparam1(visual_info);
    CAMLlocal1(dat);

    XVisualInfo * vi = XVisualInfo_val(visual_info);

    dat = caml_alloc(8, 0);
    Store_field( dat, 0, Val_Visual(vi->visual) );
    Store_field( dat, 1, Val_VisualID(vi->visualid) );
    Store_field( dat, 2, Val_screenNB(vi->screen) );
    Store_field( dat, 2, Val_int(vi->depth) );
    Store_field( dat, 3, Val_long(vi->red_mask) );
    Store_field( dat, 4, Val_long(vi->green_mask) );
    Store_field( dat, 5, Val_long(vi->blue_mask) );
    Store_field( dat, 6, Val_int(vi->colormap_size) );
    Store_field( dat, 7, Val_int(vi->bits_per_rgb) );
    CAMLreturn(dat);
}

CAMLprim value
ml_XFree_XVisualInfo( value visual_info )
{
    XVisualInfo * vi = XVisualInfo_val(visual_info);
    if (vi == NULL) {
        caml_invalid_argument("xFree_xVisualInfo: xVisualInfo NULL pointer");
    } else {
        XFree( vi );
        vi = NULL;
    }
    return Val_unit;
}

CAMLprim value
ml_XCreateColormap( value dpy, value win, value visual, value alloc )
{
    Colormap colormap = XCreateColormap(
        Display_val(dpy),
        Window_val(win),
        Visual_val(visual),
        ( Int_val(alloc) ? AllocAll : AllocNone)
    );
    return Val_Colormap(colormap);
}

CAMLprim value
ml_XFreeColormap( value dpy, value colormap )
{
    GET_STATUS  XFreeColormap(
        Display_val(dpy),
        Colormap_val(colormap) );
    CHECK_STATUS(XFreeColormap,1);
    return Val_unit;
}

CAMLprim value
ml_XSetWindowColormap( value dpy, value win, value colormap )
{
    GET_STATUS  XSetWindowColormap(
        Display_val(dpy),
        Window_val(win),
        Colormap_val(colormap) );
    CHECK_STATUS(XSetWindowColormap,1);
    return Val_unit;
}

CAMLprim value
_ml_XSetWindowAttributes_alloc( value unit )
{
    CAMLparam0();
    CAMLlocal2(ret, wattr);
    alloc_XSetWindowAttributes(wattr);
    ret = caml_alloc(2, 0);
    Store_field(ret, 0, wattr );
    Store_field(ret, 1, (value) 0 );
    CAMLreturn(ret);
}

CAMLprim value
ml_XSetWindowAttributes_alloc( value unit )
{
    CAMLparam0();
    CAMLlocal1(wattr);
    alloc_XSetWindowAttributes(wattr);
    CAMLreturn(wattr);
}


#define _WIN_ATTR_SET(field_c_type, attr_field, field_mask, conv_val, ml_type) \
\
CAMLprim value \
ml_xSetWindowAttributes_set_##attr_field( value ml_wattr, value _##attr_field ) \
{ \
    XSetWindowAttributes * wattr; \
    wattr = XSetWindowAttributes_val(Field(ml_wattr,0)); \
    unsigned long mask = (unsigned long) Field(ml_wattr,1); \
    wattr->attr_field = conv_val(_##attr_field); \
    mask |= field_mask; \
    Store_field(ml_wattr, 1, Val_ulong(mask) ); \
    return Val_unit; \
}


#define WATTR_SET(field_c_type, attr_field, field_mask, conv_val, ml_type) \
\
CAMLprim value \
ml_xSetWindowAttributes_set_##attr_field( value ml_wattr, value _##attr_field ) \
{ \
    XSetWindowAttributes * wattr; \
    wattr = XSetWindowAttributes_val(ml_wattr); \
    wattr->attr_field = conv_val(_##attr_field); \
    return Val_unit; \
}

/* setting the fields of the struct XSetWindowAttributes and the associated mask */

WATTR_SET( Pixmap,        background_pixmap,     CWBackPixmap,       Pixmap_val,    pixmap          )
WATTR_SET( unsigned long, background_pixel,      CWBackPixel,        Pixel_color_val, uint          )
WATTR_SET( Pixmap,        border_pixmap,         CWBorderPixmap,     Pixmap_val,    pixmap          )
WATTR_SET( unsigned long, border_pixel,          CWBorderPixel,      Pixel_color_val, uint          )
WATTR_SET( int,           bit_gravity,           CWBitGravity,       Int_val,       int             )
WATTR_SET( int,           win_gravity,           CWWinGravity,       Int_val,       int             )
WATTR_SET( int,           backing_store,         CWBackingStore,     Int_val,       int             )
WATTR_SET( unsigned long, backing_planes,        CWBackingPlanes,    ULong_val,     uint            )
WATTR_SET( unsigned long, backing_pixel,         CWBackingPixel,     ULong_val,     uint            )
WATTR_SET( Bool,          save_under,            CWSaveUnder,        Bool_val,      bool            )
WATTR_SET( long,          event_mask,            CWEventMask,        Eventmask_val, event_mask_list )
WATTR_SET( long,          do_not_propagate_mask, CWDontPropagate,    Long_val,      int             )
WATTR_SET( Bool,          override_redirect,     CWOverrideRedirect, Bool_val,      bool            )
WATTR_SET( Colormap,      colormap,              CWColormap,         Colormap_val,  colormap        )
WATTR_SET( Cursor,        cursor,                CWCursor,           Cursor_val,    cursor          )

static const unsigned int window_class_table[] = {
    CopyFromParent,
    InputOutput,
    InputOnly,
};

static const unsigned long winattr_valuemask_table[] = {
    CWBackPixmap,
    CWBackPixel,
    CWBorderPixmap,
    CWBorderPixel,
    CWBitGravity,
    CWWinGravity,
    CWBackingStore,
    CWBackingPlanes,
    CWBackingPixel,
    CWOverrideRedirect,
    CWSaveUnder,
    CWEventMask,
    CWDontPropagate,
    CWColormap,
    CWCursor,
};

static inline unsigned long
winattr_valuemask_val( value mask_list )
{
    unsigned long c_mask = 0; 
    while ( mask_list != Val_emptylist )
    {
        value head = Field(mask_list, 0);
        c_mask |= winattr_valuemask_table[Long_val(head)];
        mask_list = Field(mask_list, 1);
    }
    return c_mask;
}

CAMLprim value
ml_XCreateWindow(
        value dpy, value parent,
        value x, value y,
        value width, value height,
        value border_width,
        value depth, value class, value visual,
        value valuemask, value attributes )
{
    Window win = XCreateWindow(
        Display_val(dpy),
        Window_val(parent),
        Int_val(x),
        Int_val(y),
        UInt_val(width),
        UInt_val(height),
        UInt_val(border_width),
        Int_val(depth),
        window_class_table[Long_val(class)],
        Visual_val(visual),
        winattr_valuemask_val(valuemask),
        XSetWindowAttributes_val(attributes)
    );
    return Val_Window(win);
}
CAMLprim value
ml_XCreateWindow_bytecode( value * argv, int argn )
{
    return ml_XCreateWindow( argv[0], argv[1], argv[2], argv[3], argv[4], argv[5],
                             argv[6], argv[7], argv[8], argv[9], argv[10], argv[11] );
}

CAMLprim value
ml_XResizeWindow( value dpy, value win, value width, value height )
{
    GET_STATUS  XResizeWindow(
        Display_val(dpy),
        Window_val(win),
        UInt_val(width),
        UInt_val(height) );
    CHECK_STATUS(XResizeWindow,1);
    return Val_unit;
}

CAMLprim value
ml_XLowerWindow( value dpy, value win )
{
    GET_STATUS  XLowerWindow(
        Display_val(dpy),
        Window_val(win) );
    CHECK_STATUS(XLowerWindow,1);
    return Val_unit;
}

CAMLprim value
ml_XRaiseWindow( value dpy, value win )
{
    GET_STATUS  XRaiseWindow(
        Display_val(dpy),
        Window_val(win) );
    CHECK_STATUS(XRaiseWindow,1);
    return Val_unit;
}

CAMLprim value
ml_XStoreName( value dpy, value win, value name )
{
    GET_STATUS XStoreName(  
        Display_val(dpy),
        Window_val(win),
        String_val(name) );
    CHECK_STATUS(XStoreName, 1);
    return Val_unit;
}

CAMLprim value
ml_XSelectInput( value dpy, value win, value ml_event_mask )
{
    long event_mask = Eventmask_val( ml_event_mask );
    GET_STATUS  XSelectInput(
        Display_val(dpy),
        Window_val(win),
        event_mask );
    CHECK_STATUS(XSelectInput, 1);
    return Val_unit;
}

CAMLprim value
ml_XMapWindow( value dpy, value win )
{
    GET_STATUS  XMapWindow(
        Display_val(dpy),
        Window_val(win) );
    CHECK_STATUS(XMapWindow, 1);
    return Val_unit;
}

#define Val_Atom(v) ((value)(v))
#define Atom_val(v) ((Atom)(v))
#define Atom_val_addr(v) ((Atom *)(&v))

CAMLprim value
ml_XSetWMProtocols( value dpy, value win, value protocols, value count )
{
    Status status = XSetWMProtocols(
        Display_val(dpy),
        Window_val(win),
        Atom_val_addr(protocols),
        Int_val(count) );
    CHECK_STATUS(XSetWMProtocols, 1);
    return Val_unit;
}

CAMLprim value
ml_XInternAtom( value dpy, value atom_name, value only_if_exists )
{
    Atom a = XInternAtom(
        Display_val(dpy),
        String_val(atom_name),
        Bool_val(only_if_exists) );
    return Val_Atom(a);
}

/* XSizeHints, from <X11/Xutil.h> */

CAMLprim value
ml_alloc_XSizeHints( value unit )
{
    CAMLparam0();
    CAMLlocal1(size_hints);
    alloc_XSizeHints(size_hints);
    memset(XSizeHints_val(size_hints), 0, sizeof(XSizeHints));
    CAMLreturn(size_hints);
}

CAMLprim value
ml_XSizeHints_set_USPosition( value size_hints, value _x, value _y )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= USPosition;
    sh->x = Int_val(_x);
    sh->y = Int_val(_y);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PPosition( value size_hints, value _x, value _y )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PPosition;
    sh->x = Int_val(_x);
    sh->y = Int_val(_y);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_USSize( value size_hints, value _width, value _height )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= USSize;
    sh->width = Int_val(_width);
    sh->height = Int_val(_height);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PSize( value size_hints, value _width, value _height )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PSize;
    sh->width = Int_val(_width);
    sh->height = Int_val(_height);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PMinSize( value size_hints, value _min_width, value _min_height )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PMinSize;
    sh->min_width = Int_val(_min_width);
    sh->min_height = Int_val(_min_height);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PMaxSize( value size_hints, value _max_width, value _max_height )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PMaxSize;
    sh->max_width = Int_val(_max_width);
    sh->max_height = Int_val(_max_height);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PResizeInc( value size_hints, value _width_inc, value _height_inc )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PResizeInc;
    sh->width_inc = Int_val(_width_inc);
    sh->height_inc = Int_val(_height_inc);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PBaseSize( value size_hints, value _base_width, value _base_height )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PBaseSize;
    sh->base_width = Int_val(_base_width);
    sh->base_height = Int_val(_base_height);
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PAspect( value size_hints, value min_aspect, value max_aspect )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PAspect;
    sh->min_aspect.x = Int_val(Field(min_aspect,0));
    sh->min_aspect.y = Int_val(Field(min_aspect,1));
    sh->max_aspect.x = Int_val(Field(max_aspect,0));
    sh->max_aspect.y = Int_val(Field(max_aspect,1));
    return Val_unit;
}

CAMLprim value
ml_XSizeHints_set_PWinGravity( value size_hints, value _win_gravity )
{
    XSizeHints *sh;
    sh = XSizeHints_val(size_hints);
    sh->flags |= PWinGravity;
    sh->win_gravity = Int_val(_win_gravity);
    return Val_unit;
}

CAMLprim value
ml_XSetNormalHints( value dpy, value win, value size_hints )
{
    XSizeHints* hints;
    hints = XSizeHints_val(size_hints);
    GET_STATUS  XSetNormalHints(
        Display_val(dpy),
        Window_val(win),
        hints );
    CHECK_STATUS(XSetNormalHints,1);
    return Val_unit;
}

CAMLprim value
ml_XSetStandardProperties(
        value dpy, value win,
        value window_name,
        value icon_name,
        value ml_icon_pixmap,
        value ml_argv,
        value hints )
{
    int argc = Wosize_val(ml_argv);
    char ** argv = malloc((argc+1) * sizeof(char*));
    int i;
    for (i=0; i<argc; i++) {
        value ml_arg = Field(ml_argv,i);
        int len = caml_string_length(ml_arg);
        char *arg = String_val(ml_arg);
        argv[i] = malloc((len+1) * sizeof(char));
        strncpy(argv[i], arg, len);
        argv[i][len] = '\0';
    }
    argv[argc] = NULL;

    Pixmap icon_pixmap;
    if (ml_icon_pixmap == Val_int(0)) icon_pixmap = None;  // None
    else icon_pixmap = Pixmap_val( Field(ml_icon_pixmap,0) );  // Some v

    GET_STATUS  XSetStandardProperties(
        Display_val(dpy),
        Window_val(win),
        String_val(window_name),
        String_val(icon_name),
        icon_pixmap,
        argv, argc,
        XSizeHints_val(hints) );

    for (i=0; i<argc; ++i) {
        free(argv[i]);
    }
    free(argv);

    CHECK_STATUS(XSetStandardProperties,1);
    return Val_unit;
}
CAMLprim value
ml_XSetStandardProperties_bytecode( value * argv, int argn )
{
    return ml_XSetStandardProperties( argv[0], argv[1], argv[2],
                                      argv[3], argv[4], argv[5], argv[6] );
}


CAMLprim value
ml_alloc_XGCValues( value unit )
{
    CAMLparam0();
    CAMLlocal1(gcv);
    alloc_XGCValues(gcv);
    memset(XGCValues_val(gcv), 0, sizeof(XGCValues));
    CAMLreturn(gcv);
}

CAMLprim value
ml_XGCValues_set_foreground( value ml_gcv, value foreground )
{
    XGCValues * gcv;
    gcv = XGCValues_val(ml_gcv);
    gcv->foreground = Pixel_color_val(foreground);
    return Val_unit;
}

CAMLprim value
ml_XGCValues_set_background( value ml_gcv, value background )
{
    XGCValues * gcv;
    gcv = XGCValues_val(ml_gcv);
    gcv->background = Pixel_color_val(background);
    return Val_unit;
}

static const unsigned long gc_valuemask_table[] = {
    GCFunction,
    GCPlaneMask,
    GCForeground,
    GCBackground,
    GCLineWidth,
    GCLineStyle,
    GCCapStyle,
    GCJoinStyle,
    GCFillStyle,
    GCFillRule,
    GCTile,
    GCStipple,
    GCTileStipXOrigin,
    GCTileStipYOrigin,
    GCFont,
    GCSubwindowMode,
    GCGraphicsExposures,
    GCClipXOrigin,
    GCClipYOrigin,
    GCClipMask,
    GCDashOffset,
    GCDashList,
    GCArcMode,
};

static inline unsigned long
gc_valuemask_val( value mask_list )
{
    unsigned long c_mask = 0; 
    while ( mask_list != Val_emptylist )
    {
        value head = Field(mask_list, 0);
        c_mask |= gc_valuemask_table[Long_val(head)];
        mask_list = Field(mask_list, 1);
    }
    return c_mask;
}

CAMLprim value
ml_XSetForeground( value dpy, value gc, value foreground )
{
    GET_STATUS  XSetForeground(
        Display_val(dpy),
        GC_val(gc),
        Pixel_color_val(foreground)
    );
    CHECK_STATUS(XSetForeground,1);
    return Val_unit;
}

CAMLprim value
ml_XSetBackground( value dpy, value gc, value background )
{
    GET_STATUS  XSetBackground(
        Display_val(dpy),
        GC_val(gc),
        Pixel_color_val(background)
    );
    CHECK_STATUS(XSetBackground,1);
    return Val_unit;
}

CAMLprim value
ml_XCreateGC( value dpy, value d, value valuemask, value gc_values )
{
    GC gc = XCreateGC(
        Display_val(dpy),
        Drawable_val(d),
        gc_valuemask_val(valuemask),
        XGCValues_val(gc_values) );
    return Val_GC(gc);
}

CAMLprim value
ml_XFreeGC( value dpy, value gc )
{
    GET_STATUS  XFreeGC(     
        Display_val(dpy),
        GC_val(gc)
    );
    CHECK_STATUS(XFreeGC,1);
    return Val_unit;
}

CAMLprim value
ml_XClearWindow( value dpy, value win )
{
    GET_STATUS  XClearWindow(
        Display_val(dpy),
        Window_val(win)
    );
    CHECK_STATUS(XClearWindow,1);
    return Val_unit;
}

CAMLprim value
ml_XDrawRectangle( value dpy, value d, value gc, value x, value y,
                   value width, value height )
{
    GET_STATUS  XDrawRectangle(
        Display_val(dpy),
        Drawable_val(d),
        GC_val(gc),
        Int_val(x),
        Int_val(y),
        UInt_val(width),
        UInt_val(height) );
    CHECK_STATUS(XDrawRectangle, 1);
    return Val_unit;
}
CAMLprim value
ml_XDrawRectangle_bytecode( value * argv, int argn )
{
    return ml_XDrawRectangle( argv[0], argv[1], argv[2],
                              argv[3], argv[4], argv[5], argv[6] );
}
 
CAMLprim value
ml_XDrawSegments( value dpy, value d, value gc, value ml_segments )
{
    XSegment* segments;
    int nsegments = Wosize_val(ml_segments);
    int i;
    segments = calloc(nsegments, sizeof(XSegment));
    for (i=0; i<nsegments; ++i) {
        value seg = Field(ml_segments, i);
        segments[i].x1 = Int_val(Field(seg,0));
        segments[i].y1 = Int_val(Field(seg,1));
        segments[i].x2 = Int_val(Field(seg,2));
        segments[i].y2 = Int_val(Field(seg,3));
    }
    GET_STATUS  XDrawSegments(
        Display_val(dpy),
        Drawable_val(d),
        GC_val(gc),
        segments,
        nsegments
    );
    free(segments);
    CHECK_STATUS(XDrawSegments,1);
    return Val_unit;
}

CAMLprim value
ml_XDrawString( value dpy, value d, value gc, value x, value y, value str )
{
    GET_STATUS  XDrawString(
        Display_val(dpy),
        Drawable_val(d),
        GC_val(gc),
        Int_val(x),
        Int_val(y),
        String_val(str),
        caml_string_length(str) );
    CHECK_STATUS(XDrawString, 0);
    return Val_unit;
}
CAMLprim value
ml_XDrawString_bytecode( value * argv, int argn )
{
    return ml_XDrawString( argv[0], argv[1], argv[2],
                           argv[3], argv[4], argv[5] );
}

CAMLprim value
ml_XDrawArc( value dpy, value d, value gc, value x, value y,
             value width, value height, value angle1, value angle2 )
{
    GET_STATUS  XDrawArc(    
        Display_val(dpy),
        Drawable_val(d),
        GC_val(gc),
        Int_val(x),
        Int_val(y),
        UInt_val(width),
        UInt_val(height),
        Int_val(angle1),
        Int_val(angle2) );  
    CHECK_STATUS(XDrawArc, 1);
    return Val_unit;
}
CAMLprim value
ml_XDrawArc_bytecode( value * argv, int argn )
{
    return ml_XDrawArc( argv[0], argv[1], argv[2], argv[3],
                        argv[4], argv[5], argv[6], argv[7], argv[8] );
}

CAMLprim value
ml_XFillArc( value dpy, value d, value gc, value x, value y,
             value width, value height, value angle1, value angle2 )
{
    GET_STATUS  XFillArc(    
        Display_val(dpy),
        Drawable_val(d),
        GC_val(gc),
        Int_val(x),
        Int_val(y),
        UInt_val(width),
        UInt_val(height),
        Int_val(angle1),
        Int_val(angle2) );  
    CHECK_STATUS(XFillArc, 1);
    return Val_unit;
}
CAMLprim value
ml_XFillArc_bytecode( value * argv, int argn )
{
    return ml_XFillArc( argv[0], argv[1], argv[2], argv[3],
                        argv[4], argv[5], argv[6], argv[7], argv[8] );
}

CAMLprim value
ml_XFillRectangle( value dpy, value d, value gc, value x, value y,
                   value width, value height )
{
    GET_STATUS  XFillRectangle(
        Display_val(dpy),
        Drawable_val(d),
        GC_val(gc),
        Int_val(x),
        Int_val(y),
        UInt_val(width),
        UInt_val(height) );
    CHECK_STATUS(XFillRectangle, 1);
    return Val_unit;
}
CAMLprim value
ml_XFillRectangle_bytecode( value * argv, int argn )
{
    return ml_XFillRectangle( argv[0], argv[1], argv[2],
                              argv[3], argv[4], argv[5], argv[6] );
}

CAMLprim value
ml_XCopyArea( value dpy,
        value src,
        value dest,
        value gc,
        value src_x,
        value src_y,
        value width,
        value height,
        value dest_x,
        value dest_y )
{
    GET_STATUS  XCopyArea(
        Display_val(dpy),
        Drawable_val(src),
        Drawable_val(dest),
        GC_val(gc),
        Int_val(src_x),
        Int_val(src_y),
        UInt_val(width),
        UInt_val(height),
        Int_val(dest_x),
        Int_val(dest_y) );
    CHECK_STATUS(XCopyArea,1);
    return Val_unit;
}
CAMLprim value
ml_XCopyArea_bytecode( value * argv, int argn )
{
    return ml_XCopyArea( argv[0], argv[1], argv[2], argv[3], argv[4],
                         argv[5], argv[6], argv[7], argv[8], argv[9] );
}

CAMLprim value
ml_XCreatePixmap( value dpy,
        value dbl,
        value width,
        value height,
        value depth )
{
    Pixmap pixmap = XCreatePixmap(
        Display_val(dpy),
        Drawable_val(dbl),
        UInt_val(width),
        UInt_val(height),
        UInt_val(depth) );
    return Val_Pixmap(pixmap);
}


CAMLprim value
ml_XFreePixmap( value dpy, value pixmap )
{
    GET_STATUS  XFreePixmap(
        Display_val(dpy),
        Pixmap_val(pixmap)
    );
    CHECK_STATUS(XFreePixmap,1);
    return Val_unit;
}

/* Font */

CAMLprim value
ml_XLoadFont( value dpy, value name )
{
    Font font = XLoadFont(
        Display_val(dpy),
        String_val(name) );
    return Val_Font(font);
}

/* XEvent */

CAMLprim value
ml_alloc_XEvent( value unit )
{
    CAMLparam0();
    CAMLlocal1(event);
    alloc_XEvent(event);
    memset(XEvent_val(event), 0, sizeof(XEvent));
    CAMLreturn(event);
}

CAMLprim value
ml_XNextEvent( value dpy, value event )
{
    GET_STATUS  XNextEvent(
        Display_val(dpy),
        XEvent_val(event) );
    CHECK_STATUS(XNextEvent, 0);
    return Val_unit;
}

CAMLprim value
ml_XNextEvent_fun( value dpy )
{
    CAMLparam1( dpy );
    CAMLlocal1( ml_event );
    XEvent event;
    GET_STATUS  XNextEvent(
        Display_val(dpy),
        &event );
    CHECK_STATUS(XNextEvent, 0);
    copy_XEvent( event, ml_event );
    CAMLreturn( ml_event );
}

CAMLprim value
ml_XWindowEvent( value dpy, value win, value event_mask_list )
{
    CAMLparam3( dpy, win, event_mask_list );
    CAMLlocal1( event );
    long event_mask = Eventmask_val( event_mask_list );
    XEvent event_return;
    GET_STATUS  XWindowEvent(
        Display_val(dpy),
        Window_val(win),
        event_mask,
        &event_return );
    CHECK_STATUS(XWindowEvent, 0);
    copy_XEvent(event_return, event);
    CAMLreturn( event );
}

CAMLprim value
ml_XPending( value dpy )
{
    return Val_int(XPending(
        Display_val(dpy) ));
}


static const int event_type_table[] = {
    KeyPress,
    KeyRelease,
    ButtonPress,
    ButtonRelease,
    MotionNotify,
    EnterNotify,
    LeaveNotify,
    FocusIn,
    FocusOut,
    KeymapNotify,
    Expose,
    GraphicsExpose,
    NoExpose,
    VisibilityNotify,
    CreateNotify,
    DestroyNotify,
    UnmapNotify,
    MapNotify,
    MapRequest,
    ReparentNotify,
    ConfigureNotify,
    ConfigureRequest,
    GravityNotify,
    ResizeRequest,
    CirculateNotify,
    CirculateRequest,
    PropertyNotify,
    SelectionClear,
    SelectionRequest,
    SelectionNotify,
    ColormapNotify,
    ClientMessage,
    MappingNotify,
};

CAMLprim value
ml_XCheckTypedEvent( value dpy, value event_type, value event_return )
{
    if (XCheckTypedEvent(
        Display_val(dpy),
        event_type_table[Long_val(event_type)],
        XEvent_val(event_return) ))
        return Val_true;
    else
        return Val_false;
}


static const int queued_mode_table[] = {
    QueuedAlready,
    QueuedAfterFlush,
    QueuedAfterReading
};

CAMLprim value
ml_XEventsQueued( value dpy, value mode_i )
{
    int mode = queued_mode_table[Long_val(mode_i)];
    int n = XEventsQueued(
        Display_val(dpy),
        mode );
    return Val_int(n);
}

CAMLprim value
ml_XEvent_type( value event )
{
    switch (XEvent_val(event)->type)
    {
        case KeyPress:         return Val_int(0);
        case KeyRelease:       return Val_int(1);
        case ButtonPress:      return Val_int(2);
        case ButtonRelease:    return Val_int(3);
        case MotionNotify:     return Val_int(4);
        case EnterNotify:      return Val_int(5);
        case LeaveNotify:      return Val_int(6);
        case FocusIn:          return Val_int(7);
        case FocusOut:         return Val_int(8);
        case KeymapNotify:     return Val_int(9);
        case Expose:           return Val_int(10);
        case GraphicsExpose:   return Val_int(11);
        case NoExpose:         return Val_int(12);
        case VisibilityNotify: return Val_int(13);
        case CreateNotify:     return Val_int(14);
        case DestroyNotify:    return Val_int(15);
        case UnmapNotify:      return Val_int(16);
        case MapNotify:        return Val_int(17);
        case MapRequest:       return Val_int(18);
        case ReparentNotify:   return Val_int(19);
        case ConfigureNotify:  return Val_int(20);
        case ConfigureRequest: return Val_int(21);
        case GravityNotify:    return Val_int(22);
        case ResizeRequest:    return Val_int(23);
        case CirculateNotify:  return Val_int(24);
        case CirculateRequest: return Val_int(25);
        case PropertyNotify:   return Val_int(26);
        case SelectionClear:   return Val_int(27);
        case SelectionRequest: return Val_int(28);
        case SelectionNotify:  return Val_int(29);
        case ColormapNotify:   return Val_int(30);
        case ClientMessage:    return Val_int(31);
        case MappingNotify:    return Val_int(32);
        default: caml_failwith("xEventType: unhandled event type");
    }
    return Val_unit;
}


#define CHECK_EVENT_TYPE 1
#define CHECK_INT_OVERFLOW 1

CAMLprim value
ml_xKeyEvent_datas( value event )
{
    CAMLparam1( event );
    CAMLlocal1( dat );
    XEvent * e;
    e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    // The type inference should garanty this:
    if (e->type != KeyPress &&
        e->type != KeyRelease)
      caml_invalid_argument("not a key event");
#endif
#if CHECK_INT_OVERFLOW
    if (e->xkey.time > Max_long)
        caml_failwith("xKeyEvent_datas: time value overflow");
#endif
    dat = caml_alloc(8, 0);
    Store_field( dat, 0, Val_int(e->xkey.time) ); // Time: unsigned long / CARD32
    Store_field( dat, 1, Val_int(e->xkey.x) );
    Store_field( dat, 2, Val_int(e->xkey.y) );
    Store_field( dat, 3, Val_int(e->xkey.x_root) );
    Store_field( dat, 4, Val_int(e->xkey.y_root) );
    Store_field( dat, 5, Val_int(e->xkey.state) );  // ShiftMask
    Store_field( dat, 6, Val_int(e->xkey.keycode) );
    Store_field( dat, 7, Val_bool(e->xkey.same_screen) );
    CAMLreturn( dat );
}

CAMLprim value
ml_xConfigureEvent_datas( value event )
{
    CAMLparam1( event );
    CAMLlocal1( dat );
    XEvent * e;
    e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    // The type inference should garanty this:
    if (e->type != ConfigureNotify)
      caml_invalid_argument("not a ConfigureNotify event");
#endif
    dat = caml_alloc(7, 0);
    Store_field( dat, 0, Val_int(e->xconfigure.x) );
    Store_field( dat, 1, Val_int(e->xconfigure.y) );
    Store_field( dat, 2, Val_int(e->xconfigure.width) );
    Store_field( dat, 3, Val_int(e->xconfigure.height) );
    Store_field( dat, 4, Val_int(e->xconfigure.border_width) );
    Store_field( dat, 5, Val_Window(e->xconfigure.above) );
    Store_field( dat, 6, Val_bool(e->xconfigure.override_redirect) );
    CAMLreturn( dat );
}

CAMLprim value
ml_xMotionEvent_datas( value event )
{
    CAMLparam1( event );
    CAMLlocal1( dat );
    XEvent * e;
    e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    // The type inference should garanty this:
    if (e->type != MotionNotify)
      caml_invalid_argument("not a MotionNotify event");
#endif
#if CHECK_INT_OVERFLOW
    if (e->xmotion.time > Max_long)
        caml_failwith("xMotionEvent_datas: time value overflow");
#endif
    dat = caml_alloc(10, 0);
    Store_field( dat, 0, Val_Window(e->xmotion.root) );
    Store_field( dat, 1, Val_Window(e->xmotion.subwindow) );
    Store_field( dat, 2, Val_int(e->xmotion.time) ); // Time: unsigned long / CARD32
    Store_field( dat, 3, Val_int(e->xmotion.x) );
    Store_field( dat, 4, Val_int(e->xmotion.y) );
    Store_field( dat, 5, Val_int(e->xmotion.x_root) );
    Store_field( dat, 6, Val_int(e->xmotion.y_root) );
    Store_field( dat, 7, Val_uint(e->xmotion.state) );
    Store_field( dat, 8, Val_char(e->xmotion.is_hint) );
    Store_field( dat, 9, Val_bool(e->xmotion.same_screen) );
    CAMLreturn( dat );
}

static const unsigned int button_table[] = {
    AnyButton,
    Button1,
    Button2,
    Button3,
    Button4,
    Button5,
};
    //unsigned int button = button_table[Long_val(v)];
#define Button_val(v) (button_table[Long_val(v)])

static inline value Val_button(unsigned int b)
{
    switch (b) {
        case AnyButton: return Val_int(0);
        case Button1: return Val_int(1);
        case Button2: return Val_int(2);
        case Button3: return Val_int(3);
        case Button4: return Val_int(4);
        case Button5: return Val_int(5);
    }
    return Val_int(0);
}

CAMLprim value
ml_xButtonEvent_datas( value event )
{
    CAMLparam1( event );
    CAMLlocal1( dat );
    XEvent * e;
    e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    // The type inference should garanty this:
    if (e->type != ButtonPress &&
        e->type != ButtonRelease)
      caml_invalid_argument("not a MotionNotify event");
#endif
#if CHECK_INT_OVERFLOW
    if (e->xmotion.time > Max_long)
        caml_failwith("xButtonEvent_datas: time value overflow");
#endif
    dat = caml_alloc(10, 0);
    Store_field( dat, 0, Val_Window(e->xbutton.root) );
    Store_field( dat, 1, Val_Window(e->xbutton.subwindow) );
    Store_field( dat, 2, Val_int(e->xbutton.time) ); // Time: unsigned long / CARD32
    Store_field( dat, 3, Val_int(e->xbutton.x) );
    Store_field( dat, 4, Val_int(e->xbutton.y) );
    Store_field( dat, 5, Val_int(e->xbutton.x_root) );
    Store_field( dat, 6, Val_int(e->xbutton.y_root) );
    Store_field( dat, 7, Val_uint(e->xbutton.state) );
    Store_field( dat, 8, Val_button(e->xbutton.button) );
    Store_field( dat, 9, Val_bool(e->xbutton.same_screen) );
    CAMLreturn( dat );
}

CAMLprim value
ml_xEvent_xclient_data_l_0( value event )
{
    XEvent * e;
    e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    // The type inference should garanty this:
    if (e->type != ClientMessage)
      caml_invalid_argument("not a ClientMessage event");
#endif
    long atom = e->xclient.data.l[0];
    return Val_Atom( atom );
}

CAMLprim value
ml_XLookupString( value event, value buffer )
{
    CAMLparam1( event );
    CAMLlocal1( ret );
    KeySym keysym;
    XEvent * e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    if (e->type != KeyPress &&
        e->type != KeyRelease)
      caml_invalid_argument("not a key event");
#endif
    int nchars = XLookupString(  // from <X11/Xutil.h>
        &(e->xkey),
        String_val(buffer),
        caml_string_length(buffer),
        &keysym,
        NULL );
    ret = caml_alloc(2, 0);
    Store_field( ret, 0, Val_int(nchars) );
    Store_field( ret, 1, Val_keysym(keysym) );
    CAMLreturn( ret );
}

CAMLprim value
ml_XLookupKeysym( value event, value index )
{
    XEvent * e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    if (e->type != KeyPress &&
        e->type != KeyRelease)
      caml_invalid_argument("not a key event");
#endif
    KeySym keysym = XLookupKeysym(
        &(e->xkey),
        Int_val(index) );  // index from 0 to pred keysyms_per_keycode
    return Val_keysym(keysym);
}

CAMLprim value
ml_XKeycodeToKeysym( value dpy, value keycode, value index )
{
    KeySym keysym = XKeycodeToKeysym(
        Display_val(dpy),
        ULong_val(keycode),
        Int_val(index) );
    return Val_keysym(keysym);
}

CAMLprim value
ml_XRefreshKeyboardMapping( value event )
{
    XEvent *e;
    e = XEvent_val(event);
#if CHECK_EVENT_TYPE
    if (e->type != MappingNotify)
      caml_invalid_argument("not a MappingNotify event");
#endif
    XRefreshKeyboardMapping(&(e->xmapping));
    return Val_unit;
}


/* Keyboard */

CAMLprim value
ml_XAutoRepeatOff( value dpy )
{
    GET_STATUS XAutoRepeatOff( Display_val(dpy) );
    CHECK_STATUS(XAutoRepeatOff,1);
    return Val_unit;
}

CAMLprim value
ml_XAutoRepeatOn( value dpy )
{
    GET_STATUS XAutoRepeatOn( Display_val(dpy) );
    CHECK_STATUS(XAutoRepeatOn,1);
    return Val_unit;
}

CAMLprim value
ml_XQueryKeymap( value dpy )
{
    CAMLparam1( dpy );
    CAMLlocal1( ml_keys );
    ml_keys = caml_alloc_string(32);
    char *keys_ptr;
    keys_ptr = String_val(ml_keys);
    GET_STATUS XQueryKeymap(
        Display_val(dpy),
        keys_ptr );
    CHECK_STATUS(XQueryKeymap,1);
    CAMLreturn( ml_keys );
}

CAMLprim value
ml_XGetKeyboardControl( value dpy )
{
    CAMLparam1( dpy );
    CAMLlocal2( tpl, ml_auto_repeats );
    XKeyboardState kbs;
    GET_STATUS  XGetKeyboardControl(
        Display_val(dpy),
        &kbs );
    CHECK_STATUS(XGetKeyboardControl,1);
    tpl = caml_alloc(7, 0);
    ml_auto_repeats = caml_alloc_string(32);
    memcpy(String_val(ml_auto_repeats), kbs.auto_repeats, 32);
    Store_field( tpl, 0, Val_int(kbs.key_click_percent) );
    Store_field( tpl, 1, Val_int(kbs.bell_percent) );
    Store_field( tpl, 2, Val_uint(kbs.bell_pitch) );
    Store_field( tpl, 3, Val_uint(kbs.bell_duration) );
    Store_field( tpl, 4, Val_ulong(kbs.led_mask) );
    Store_field( tpl, 5, Val_int( (kbs.global_auto_repeat == AutoRepeatModeOff ? 0 : 1)) );
    Store_field( tpl, 6, ml_auto_repeats );
    CAMLreturn( tpl );
}


/* ScreenSaver */

CAMLprim value
ml_XForceScreenSaver( value dpy, value mode )
{
    GET_STATUS  XForceScreenSaver(
        Display_val(dpy),
        ( Int_val(mode) ? ScreenSaverReset : ScreenSaverActive )
    );
    CHECK_STATUS(XForceScreenSaver,1);
    return Val_unit;
}

CAMLprim value
ml_XGetScreenSaver( value dpy )
{
    CAMLparam1( dpy );
    CAMLlocal1( tpl );
    int timeout;
    int interval;
    int prefer_blanking;
    int allow_exposures;
    GET_STATUS  XGetScreenSaver(
        Display_val(dpy),
        &timeout,
        &interval,
        &prefer_blanking,
        &allow_exposures
    );
    CHECK_STATUS(XGetScreenSaver,1);
    tpl = caml_alloc(4, 0);
    Store_field( tpl, 0, Val_int(timeout) );
    Store_field( tpl, 1, Val_int(interval) );
    Store_field( tpl, 2, Val_int(prefer_blanking) );
    Store_field( tpl, 3, Val_int(allow_exposures) );
    CAMLreturn( tpl );
}

/* {{{ (Screen *) */

CAMLprim value
ml_XDefaultScreenOfDisplay( value dpy )
{
    Screen *scr = _xDefaultScreenOfDisplay(
        Display_val(dpy) );
    return Val_XScreen(scr);
}

CAMLprim value
ml_XScreenOfDisplay( value dpy, value screen_number )
{
    Screen *scr = _xScreenOfDisplay(
        Display_val(dpy),
        ScreenNB_val(screen_number) );
    return Val_XScreen(scr);
}

CAMLprim value
ml_XDefaultVisualOfScreen( value xscreen )
{
    Visual *visual = _xDefaultVisualOfScreen(
        XScreen_val(xscreen) );
    return Val_Visual(visual);
}

CAMLprim value
ml_XRootWindowOfScreen( value xscreen )
{
    return Val_Window( _xRootWindowOfScreen( XScreen_val(xscreen) ));
}

CAMLprim value
ml_XBlackPixelOfScreen( value xscreen )
{
    unsigned long px = _xBlackPixelOfScreen(
        XScreen_val(xscreen)
    );
    return Val_pixel_color(px);
}

CAMLprim value
ml_XWhitePixelOfScreen( value xscreen )
{
    unsigned long px = _xWhitePixelOfScreen(
        XScreen_val(xscreen)
    );
    return Val_pixel_color(px);
}

CAMLprim value
ml_XDefaultColormapOfScreen( value xscreen )
{
    Colormap cmap = _xDefaultColormapOfScreen(
        XScreen_val(xscreen)
    );
    return Val_Colormap(cmap);
}

CAMLprim value
ml_XDefaultDepthOfScreen( value xscreen )
{
    return Val_int( _xDefaultDepthOfScreen( XScreen_val(xscreen) ));
}

CAMLprim value
ml_XDefaultGCOfScreen( value xscreen )
{
    GC gc = _xDefaultGCOfScreen(
        XScreen_val(xscreen)
    );
    return Val_GC(gc);
}

CAMLprim value
ml_XWidthOfScreen( value xscreen )
{
    return Val_int( _xWidthOfScreen( XScreen_val(xscreen) ));
}

CAMLprim value
ml_XHeightOfScreen( value xscreen )
{
    return Val_int( _xHeightOfScreen( XScreen_val(xscreen) ));
}

/* }}} */

// vim: sw=4 sts=4 ts=4 et fdm=marker
